
-- Maze generation functions
local M = {}

M.wall = 0
M.path = 1

local function isValidDirection(x, y, d, notes)
    -- check for out of bounds
    if x + d[1] * 2 > 1 and x + d[1] * 2 < notes.w and y + d[2] * 2 > 1 and y + d[2] * 2 < notes.h then
        -- check if visited
        if not notes[x + d[1] * 2][y + d[2] * 2].visited then
            return true
        end
    end
    return false
end

function M.newMaze(w, h)
    -- an implementation of Recursive Backtracking
    -- based on https://weblog.jamisbuck.org/2010/12/27/maze-generation-recursive-backtracking
    
-- fixme: BUG: y sometimes becomes nil
-- fixme: BUY: backtrack sometimes goes back TWO steps at a time
    
    -- initialize parameters
    w = w or 51
    h = h or 41
    if math.floor(w / 2) * 2 == w then w = w + 1 end -- width and height must be odd numbers
    if math.floor(h / 2) * 2 == h then h = h + 1 end
    player.x, player.y = 2, math.floor(love.math.random(2, h) / 2) * 2 -- start on an even cell
    local directions = {{0, -1}, {0, 1}, {-1, 0}, {1, 0}} -- direction deltas
    local maze = {w=w, h=h}  -- the maze to return when finished
    local notes = {w=w, h=h} -- backtrack etc notes
    
    -- initialize maze & notes grids
    for x = 1, w do
        maze[x] = {}
        notes[x] = {}
        for y = 1, h do
            maze[x][y] = M.wall
            notes[x][y] = {}
        end
    end
    
    -- initialize maze
    local x, y = player.x, player.y
    maze[x][y] = M.path
    notes[x][y].visited = true
    
    -- DEBUG INFO --
    print()
    print("Creating maze "..w.." x "..h)
    print("Starting at "..x..","..y)
    -- DEBUG INFO --
    
    -- maze loop
    while true do
        
        -- check for valid directions
        local valid
        repeat
            valid = isValidDirection(x, y, directions[1], notes) or
                    isValidDirection(x, y, directions[2], notes) or
                    isValidDirection(x, y, directions[3], notes) or
                    isValidDirection(x, y, directions[4], notes)
            if not valid then
                
                -- DEBUG INFO --
                print("Backtrack target: "..notes[x][y].x0..","..notes[x][y].y0) -- here's proof these are valid entries...
                -- DEBUG INFO --
                
                x = notes[x][y].x0 -- x is never affected by this bug...
                y = notes[x][y].y0 -- somehow this can set y to nil even though I've just shown a non-nil value for y0...
                
                
                -- DEBUG INFO --
                if y == nil then
                    print("ERROR!    x == ", x, "  y == ", y) -- only y is affected by this bug...
                    print("Started at "..player.x..","..player.y)
                    return maze
                end
                print("Returning to "..x..","..y)
                -- DEBUG INFO --
                
            end
        until valid or (x == player.x and y == player.y)
        
        -- check if maze is finished
        if not valid then break end
        
        -- pick a valid direction
        local d = love.math.random(1, 4)
        while not isValidDirection(x, y, directions[d], notes) do
            d = d + 1
            if d > 4 then d = 1 end
        end
        d = directions[d]
        
        -- dig two steps in that direction
        local x1, y1 = x  + d[1], y  + d[2]
        local x2, y2 = x1 + d[1], y1 + d[2]
        maze[x1][y1] = M.path
        maze[x2][y2] = M.path
        notes[x1][y1].visited = true
        notes[x2][y2].visited = true
        notes[x2][y2].x0 = x
        notes[x2][y2].y0 = y
        if y == nil or notes[x2][y2].y0 == nil then print("Assignment error!  x == ", x, "  y == ", y) end
        x = x2
        y = y2
        
        -- DEBUG INFO --
        print("Dug to "..x2..","..y2)
        -- DEBUG INFO --
    end
    
    -- done!
print("done!")
    return maze
end

return M
